# nms_3d/plot_3d_boxes.py

import os
import torch
import plotly.graph_objects as go
from typing import Union, Tuple

def plot_3d_boxes(boxes: torch.Tensor, 
                  title: str = "Plot 3D boxes", 
                  save_html: bool = False, 
                  html_filename_path: Union[str, None] = "./plot_3d_boxes.html",
                  color: Tuple[int, int, int] = (255, 0, 0),
                  show_scores: bool = True) -> None:
    """
    Save a 3D plot with the bounding boxes and optionally display each box's score as a label.

    :param boxes: tensor containing 3D bounding box coordinates with columns 'SCORE', 'X MIN', 'Y MIN', 'Z MIN', 'X MAX', 'Y MAX', 'Z MAX'.
    :param title: title of the plot. Default is "Plot 3D boxes".
    :param save_html: whether to save the plot in an HTML file. Default is False.
    :param html_filename_path: name of the HTML file to save. Default is "./plot_3d_boxes.html".
    :param color: RGB color tuple for the boxes. Default is (255, 0, 0) for red.
    :param show_scores: whether to display the scores on the boxes. Default is True.

    Example:

        import torch
        from nms_3d import plot_3d_boxes

        # tensor containing 3D bounding boxes with confidence scores
        boxes = torch.tensor([
            [0.95, 10, 10, 10, 20, 20, 20],  # high-score box
            [0.85, 30, 30, 30, 40, 40, 40],  # another box
            [0.75, 50, 50, 50, 60, 60, 60]   # third box
        ])

        # set the parameters for visualization
        title = "3D Bounding Box Visualization"
        save_html = True  # set to True if you want to save as HTML
        html_filename_path = "./bounding_boxes_plot.html"
        color = (0, 255, 0)  # green color for the boxes
        show_scores = True  # display scores on the boxes

        # call the function to plot the boxes
        plot_3d_boxes(boxes=boxes, 
                      title=title, 
                      save_html=save_html, 
                      html_filename_path=html_filename_path, 
                      color=color, 
                      show_scores=show_scores)

    """

    # validate the input tensor
    if not isinstance(boxes, torch.Tensor):
        raise TypeError(f"Expected 'boxes' to be a torch.Tensor, but got {type(boxes)} instead.")
    if boxes.ndim != 2 or boxes.size(1) != 7:
        raise ValueError(f"'boxes' should have shape (N, 7) but got {boxes.shape}.")

    # validate HTML filename if saving to HTML
    if save_html:
        if not isinstance(html_filename_path, str) or not html_filename_path.endswith(".html"):
            raise ValueError("Invalid HTML filename. Please provide a valid path with '.html' extension.")

    # unpack color and set up with desired alpha values
    color_border = f'rgba({color[0]}, {color[1]}, {color[2]}, 1)'
    color_face = f'rgba({color[0]}, {color[1]}, {color[2]}, 0.5)'

    fig = go.Figure()

    for i in range(boxes.size(0)):  # iterate through rows in the tensor
        box = boxes[i]

        # calculate the centroid of the box for positioning the score label
        centroid = [(box[1] + box[4]) / 2, (box[2] + box[5]) / 2, (box[3] + box[6]) / 2]

        vertices = [
            [box[1], box[2], box[3]],  # X MIN, Y MIN, Z MIN
            [box[4], box[2], box[3]],  # X MAX, Y MIN, Z MIN
            [box[4], box[5], box[3]],  # X MAX, Y MAX, Z MIN
            [box[1], box[5], box[3]],  # X MIN, Y MAX, Z MIN
            [box[1], box[2], box[6]],  # X MIN, Y MIN, Z MAX
            [box[4], box[2], box[6]],  # X MAX, Y MIN, Z MAX
            [box[4], box[5], box[6]],  # X MAX, Y MAX, Z MAX
            [box[1], box[5], box[6]],  # X MIN, Y MAX, Z MAX
        ]

        # connect vertices to form box
        vertices.append(vertices[0])  # close the box

        # draw the border of the six faces
        #                 
        #     6 • - - - - • 5
        #      /|        /|
        #     / |       / |
        #  7 • - - - - • 4|
        #    |  |      |  |
        #    |  |      |  | 
        #    |2 • - - -|- • 1
        #    | /       | /
        #    |/        |/
        #  3 • - - - - • 0
        # 
        faces = [
            [vertices[0], vertices[1], vertices[2], vertices[3], vertices[0]],
            [vertices[4], vertices[5], vertices[6], vertices[7], vertices[4]],
            [vertices[0], vertices[1], vertices[5], vertices[4], vertices[0]],
            [vertices[2], vertices[3], vertices[7], vertices[6], vertices[2]],
            [vertices[1], vertices[2], vertices[6], vertices[5], vertices[1]],
            [vertices[3], vertices[0], vertices[4], vertices[7], vertices[3]],
        ]

        for face in faces:
            fx, fy, fz = zip(*face)

            # draw lines connecting vertices to form faces
            fig.add_trace(go.Scatter3d(
                x=fx,
                y=fy,
                z=fz,
                mode='lines',
                line=dict(color=color_border, width=2),
                showlegend=False,
            ))

        # add a trace for the filled faces
        triangular_face = [
            [0, 1, 2], [0, 2, 3], [4, 5, 6], [4, 6, 7], 
            [2, 3, 7], [2, 6, 7], [0, 1, 4], [1, 4, 5], 
            [0, 3, 7], [0, 4, 7], [1, 2, 6], [1, 5, 6]
        ]

        fig.add_trace(go.Mesh3d(
            x=[v[0] for v in vertices],
            y=[v[1] for v in vertices],
            z=[v[2] for v in vertices],
            i=[face[0] for face in triangular_face],
            j=[face[1] for face in triangular_face],
            k=[face[2] for face in triangular_face],
            opacity=0.5,
            color=color_face
        ))

        # conditionally add the score label at the centroid
        if show_scores:
            fig.add_trace(go.Scatter3d(
                x=[centroid[0]],
                y=[centroid[1]],
                z=[centroid[2]],
                mode='text',
                text=f'{box[0]:.2f}',
                textposition="middle center",
                showlegend=False
            ))

    # set axis labels and plot title
    fig.update_layout(
        scene=dict(
            xaxis_title='X-axis',
            yaxis_title='Y-axis',
            zaxis_title='Z-axis',
        ),
        title=title,
    )

    if save_html:
        html_filename_path = os.path.abspath(html_filename_path)
        fig.write_html(html_filename_path)
        print(f"Plot saved as HTML: {html_filename_path}")
    else:
        fig.show()



# nms_3d/nms_3d.py

import torch
import logging


def nms_3d(prediction_boxes: torch.Tensor, 
           iou_threshold: float = 0.5, 
           debug: bool = False) -> torch.Tensor:
    """
    Perform 3D Non-Maximum Suppression on a set of bounding boxes.
    
    :param prediction_boxes: Tensor of shape (N, 7), where N is the number of bounding boxes. 
                             Each row should be of format: 'SCORE', 'X MIN', 'Y MIN', 'Z MIN', 'X MAX', 'Y MAX', 'Z MAX'.
    :param iou_threshold: Intersection over union threshold between 0 and 1. Default is 0.5.
    :param debug: Verbose print about the boxes suppression. Default is False.
    
    :return: Tensor containing the selected bounding boxes after applying NMS.

    Example:

        import torch
        from nms_3d import nms_3d

        # tensor of 3D bounding boxes with scores
        prediction_boxes = torch.tensor([
            [0.95, 10, 10, 10, 20, 20, 20],  # high-score box
            [0.90, 12, 12, 12, 22, 22, 22],  # overlapping box (should be suppressed)
            [0.85, 50, 50, 50, 60, 60, 60],  # distant box (should be kept)
            [0.80, 55, 55, 55, 65, 65, 65],  # overlapping with previous (should be suppressed)
            [0.75, 100, 100, 100, 110, 110, 110]  # another distant box (should be kept)
        ])

        # define the IoU threshold
        iou_threshold = 0.5

        # run the 3D Non-Maximum Suppression
        filtered_boxes = nms_3d(prediction_boxes=prediction_boxes,
                                iou_threshold=iou_threshold,
                                debug=True)

    """

    # configure logging
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # input validation
    if not isinstance(prediction_boxes, torch.Tensor):
        prediction_boxes = torch.tensor(prediction_boxes)

    if prediction_boxes.ndim != 2 or prediction_boxes.size(1) != 7:
        raise ValueError(f"prediction_boxes must be of shape (N, 7) formatted as: 'SCORE', 'X MIN', 'Y MIN', 'Z MIN', 'X MAX', 'Y MAX', 'Z MAX'. Got {prediction_boxes.shape} instead")

    # check iou value
    if not (0 <= iou_threshold <= 1):
        raise ValueError(f"iou_threshold must be between 0 and 1. Got {iou_threshold} instead")


    best_boxes_hist = []

    while prediction_boxes.size(0) > 0:
        # sort predictions by score in descending order
        _, sorted_indices = torch.sort(prediction_boxes[:, 0], descending=True)
        prediction_boxes = prediction_boxes[sorted_indices]

        highest_score_box = prediction_boxes[0]

        # calculate intersection coordinates
        x_min_intersection = torch.max(highest_score_box[1], prediction_boxes[:, 1])
        y_min_intersection = torch.max(highest_score_box[2], prediction_boxes[:, 2])
        z_min_intersection = torch.max(highest_score_box[3], prediction_boxes[:, 3])
        x_max_intersection = torch.min(highest_score_box[4], prediction_boxes[:, 4])
        y_max_intersection = torch.min(highest_score_box[5], prediction_boxes[:, 5])
        z_max_intersection = torch.min(highest_score_box[6], prediction_boxes[:, 6])

        # calculate intersection volume
        intersection_volume = torch.max(torch.tensor(0.0), x_max_intersection - x_min_intersection) * \
                              torch.max(torch.tensor(0.0), y_max_intersection - y_min_intersection) * \
                              torch.max(torch.tensor(0.0), z_max_intersection - z_min_intersection)

        # calculate volumes
        highest_score_box_volume = (highest_score_box[4] - highest_score_box[1]) * \
                                   (highest_score_box[5] - highest_score_box[2]) * \
                                   (highest_score_box[6] - highest_score_box[3])

        row_volume = (prediction_boxes[:, 4] - prediction_boxes[:, 1]) * \
                     (prediction_boxes[:, 5] - prediction_boxes[:, 2]) * \
                     (prediction_boxes[:, 6] - prediction_boxes[:, 3])

        # calculate iou values
        iou_values = intersection_volume / (highest_score_box_volume + row_volume - intersection_volume)

        # create a mask to threshold over the boxes that need to be suppressed
        iou_threshold_mask = iou_values > iou_threshold

        # select the rows with iou > threshold
        iou_threshold_boxes = prediction_boxes[iou_threshold_mask]

        # save the highest score box into the best boxes list
        best_boxes_hist.append(iou_threshold_boxes[0])

        # debug print
        if debug:
            print(f"Highest Score Box: SCORE={highest_score_box[0]:.4f} / BOX={highest_score_box[1:].tolist()}")
            for i in range(len(iou_threshold_boxes)):
                suppressed_box = iou_threshold_boxes[i]
                if (int(iou_values[i]) != 1): # exclude from debug print the intersection between the highest score and itself
                    print(f"Suppressed Box: SCORE={suppressed_box[0]:.4f} / BOX={suppressed_box[1:].tolist()} / IoU={iou_values[i]:.4f}")
            print("-" * 100)

        # remove threshold boxes
        prediction_boxes = prediction_boxes[~iou_threshold_mask]

    return torch.stack(best_boxes_hist)


from setuptools import setup, find_packages

setup(
    name="nms_3d",
    version="0.1.6",
    description="3D bounding box NMS and plotting package",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Giulio Russo, Ciro Russo",
    author_email="russogiulio1998@icloud.com, ciro.russo2910@gmail.com",
    url="https://github.com/GiulioRusso/NMS-3D",
    packages=find_packages(),
    license="MIT License",
    install_requires=[
        "torch>=2.2.2",
        "plotly>=5.13.1"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9.5",
)
